#include <stdio.h>
#include <stdlib.h>  // for atoi(3)
#include <limits.h>  // for INT_MAX
#include <string.h>  // for strcmp() 
#include <unistd.h>  // for usleep() 
#include <math.h>    // for pow, sin, cos
#include <libplayerc++/playerc++.h>  


#define USAGE \
  "USAGE: ipr-proj2      [-a <avoidPast>]\n" \
  "       -a<avoidPast>: use avoidPast schema\n" \


#ifndef MIN
#define MIN(a,b) ((a < b) ? (a) (b))
#endif
#ifndef MAX
#define MAX(a,b) ((a > b) ? (a) (b))
#endif

char   host[256] = "localhost";
int    port = 6665;
int    index_position = 0;  // use this to access the position devices
int    index_camera_L = 0;  // use this to access the left  camera devices
int    index_camera_R = 1;  // use this to access the right camera devices
double focalLength = 1000.; // focal length of stereo camera, in pixel
double cameraDist  = 0.12;  // the distance between stereo cameras, in meter
double goalX = 6.0;         // x coordinate of goal postion
double goalY = 0.0;         // y coordinate of goal postion 
double initialX  = -6.0;    // x coordinate of initial position
double initialY  =  0.0;    // y coordiante of initial postion
double initialT  =  0;      // initial orietation
double object_threshold =3; //If an object is close than this threshold 
double sumX=0;
double sumY=0;
double X_obj = 0;
double Y_obj = 0;
double sum_X_obj=0;
double sum_Y_obj = 0;
double X_goal = 0;
double Y_goal = 0;
double past_X ;
double past_Y;
bool   avoidPast = false;   // if use avoid past scheme or not 


void parse_args(int argc, char** argv)
{
    int i;
    i=1;
    while(i<argc)
    {   
	if(!strcmp(argv[i],"-a"))
	{
	    if(++i<argc)
		avoidPast = atoi(argv[i]);
	    else
	    {
		puts(USAGE);
		exit(1);   
	    }
	}
        else
        {
            puts(USAGE);
	    exit(1);
	}
	i++;
    }
}


/*This function returns the pairs of corresponding pixels of the detected obstacle 
  in two cameras, also return the groundtruth of the range and bearing which you 
  could use to validate your stereo result. Student shouldn't change the content*/
void getImageCord(player_fiducial_item_t fL, player_fiducial_item_t fR, double* uL, double* uR, double* gtr, double* gtb)
{
    *uL  = (-fL.pose.py) * focalLength / fL.pose.px;
    *uR  = (-fR.pose.py) * focalLength / fR.pose.px; 
    *gtr = (pow(pow(fL.pose.px, 2)+pow(fL.pose.py, 2), 0.5)+pow(pow(fR.pose.px, 2)+pow(fR.pose.py, 2), 0.5))/2;
    *gtb = atan((fL.pose.py+cameraDist/2)/fL.pose.px);
}


int main(int argc, char **argv)
{
    using namespace PlayerCc;

    parse_args(argc, argv);
 
    /*current localization of the robot*/
    double currX, currY, currT; 

    /*the image coordinates in two cameras*/
    double uL,  uR;

    /*groundtruth of range and bearing of the obstacle relative to camera*/
    double gtr, gtb;

    /*rotation and speed used for controlling your robot*/
    double turnrate = 0;
    double curspeed = 0.3;

    /*Sets up the connections to the Player server*/
    PlayerClient    robot(host, port);
    Position2dProxy pp(&robot,index_position);
    FiducialProxy   fpL(&robot,index_camera_L); // left  camera
    FiducialProxy   fpR(&robot,index_camera_R); // right camera
 
    /*measure how far are we from the goal at the beginning*/
    double distFromGoal;
    distFromGoal = pow(pow(initialX - goalX, 2) + pow(initialY - goalY, 2), 0.5);  

    /*Sets the odometry to the pose (x, y, yaw) */
    pp.SetOdometry(initialX, initialY, initialT);

    /*stop when we're < 0.5m from the goal*/
    while (distFromGoal > 0.5)
    {
        /*read messages from the player server*/
	robot.Read();
     
	/*measure how far are we from the goal now*/
	currX = pp.GetXPos();
	currY = pp.GetYPos();
	currT = pp.GetYaw();
	distFromGoal = pow(pow(currX - goalX, 2) + pow(currY - goalY, 2), 0.5); 
        printf("distFromGoal: %f\n", distFromGoal);

        /*pair up the detected obstacles between two cameras, store them in two arrays  
          e.g. uLs[0] and uRs[0] are a pair of correpondence, you should use uLs and uRs 
          to do stereo. you could also use the groundtruth of range and bearing to 
          validate your result */
	int numL =  fpL.GetCount(); 
	//printf("numL = %d\n", numL);
	int numR =  fpR.GetCount();
        int numPair  = MIN(numL,numR);
        double* uLs  = (double*)malloc(numPair*sizeof(double));
	double* uRs  = (double*)malloc(numPair*sizeof(double));   
	double* gtrs = (double*)malloc(numPair*sizeof(double));    /*Ground Truth Ranges*/
	double* gtbs = (double*)malloc(numPair*sizeof(double));   /*Ground truth bearings*/

	int ptr = 0;
	for(int i=0; i<numL; i++ )
        {
	    for (int j=0; j<numR; j++)
	    {
		if (fpL[i].id == fpR[j].id)  /*if fudicial point left = fudicial point right*/
		{
		    getImageCord(fpL[i], fpR[j], &uL, &uR, &gtr, &gtb); /*ul location in camera plane left, location in camera plane right*/
                    uLs[ptr]  = uL;
                    uRs[ptr]  = uR;
		    gtrs[ptr] = gtr;
                    gtbs[ptr] = gtb;  
                    ptr++;   
		    break;
		}
	
	    }
	}
	//printf("x_obj %d\n", numPair);
	/*********************************Insert your code below***********************************/
     //search through gtb + math pow and gtr
	for(int i =0; i<=numPair; i++) {
		//calc vectr to robot for object i
		X_obj = 43*(gtrs[i]*cos(gtbs[i]));
		Y_obj = 43*(gtrs[i]*sin(gtbs[i]));

		sum_X_obj = sum_X_obj + X_obj;
		sum_Y_obj = sum_Y_obj + Y_obj;
		
		//printf("x_obj %f\n", X_obj);
		//printf("Y_obj %f\n", Y_obj);
		//printf("sum_X_obj %f\n", sum_X_obj);
		//printf("sum_Y_obj %f\n", sum_Y_obj);
		
	}
	printf("=============================\n");
	//calc vectoro goal
	double k = 45;
	X_goal = k*(goalX-currX)*(rand()/((double)RAND_MAX+1));
	Y_goal = k*(goalY-currY)*(rand()/((double)RAND_MAX+1));
	/*store current value in array and push*/
	
	past_X = (currX +0.5*past_X);
	past_Y = (currY + 0.9*past_Y);

	sumX = X_goal -sum_X_obj-10*past_X;
	sumY = Y_goal - sum_Y_obj-1000000*past_Y;
	//printf("Past_X %f\n" ,past_X);
	//printf("Past_Y %f\n" , past_Y);
	printf("sumX %f\n", sumX);
	printf("sumY %f\n", sumY);
	//Apply to motors	
	turnrate = 0.25*(atan(sumY/sumX) - currT) + 0.001*atan(sumY/sumX); //PD control on my wheels baby!
	//printf("turnrate %f\n",turnrate);
    /*********************************Insert your code above***********************************/

	pp.SetSpeed( curspeed, turnrate );

        free(uLs);
        free(uRs);
        free(gtrs);
	free(gtbs);

    }


    printf("**************************************\n");
    printf("WOOHOO!!!  WE MADE IT!!!  WOOOOHOOO!!!\n");
    printf("**************************************\n");
	
}
